import torch
import numpy as np
import pandas as pd
from torch import nn
from torch.utils.data import DataLoader
from pathlib import Path
from typing import Optional

from now_2023.datasets import MRIDataset

import warnings

warnings.filterwarnings("ignore")


class CNNModel:
    """The CNN model."""

    def __init__(self, learning_rate: float, n_epochs: int, batch_size: int):
        self._learning_rate = learning_rate
        self._n_epochs = n_epochs
        self._batch_size = batch_size
        self.use_cuda = torch.cuda.is_available()
        self._network = CustomNetwork()
        if self.use_cuda:
            self._network = self._network.cuda()
        self._criterion = nn.CrossEntropyLoss(reduction="sum")
        self._optimizer = torch.optim.Adam(
            self._network.parameters(), self._learning_rate
        )
        self._estimator = None

    def fit(
        self,
        img_dir: Path,
        df: pd.DataFrame,
        transform=None,
        batch_size: Optional[int] = None,
    ):
        train_dataset = MRIDataset(img_dir, df, transform=transform)
        loader = DataLoader(
            train_dataset,
            batch_size=batch_size or self._batch_size,
            shuffle=True,
            # num_workers=1,
            pin_memory=True,
        )
        self._estimator = train(
            self._network,
            loader,
            self._criterion,
            self._optimizer,
            self._n_epochs,
        )

    def predict(
        self,
        img_dir: Path,
        df: pd.DataFrame,
        transform=None,
        batch_size: Optional[int] = None,
    ):
        valid_dataset = MRIDataset(img_dir, df, transform=transform)
        loader = DataLoader(
            valid_dataset,
            batch_size=batch_size or self._batch_size,
            shuffle=False,
            # num_workers=1,
            pin_memory=True,
        )
        return test(
            self._estimator,
            loader,
            self._criterion,
        )

    def save(self, output_filename: Path) -> None:
        torch.save(self._estimator, output_filename)


class PadMaxPool3d(nn.Module):
    """A MaxPooling module which deals with odd sizes with padding"""

    def __init__(self, kernel_size, stride, return_indices=False, return_pad=False):
        super(PadMaxPool3d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.pool = nn.MaxPool3d(kernel_size, stride, return_indices=return_indices)
        self.pad = nn.ConstantPad3d(padding=0, value=0)
        self.return_indices = return_indices
        self.return_pad = return_pad

    def set_new_return(self, return_indices=True, return_pad=True):
        self.return_indices = return_indices
        self.return_pad = return_pad
        self.pool.return_indices = return_indices

    def forward(self, f_maps):
        coords = [self.stride - f_maps.size(i + 2) % self.stride for i in range(3)]
        for i, coord in enumerate(coords):
            if coord == self.stride:
                coords[i] = 0

        self.pad.padding = (coords[2], 0, coords[1], 0, coords[0], 0)

        if self.return_indices:
            output, indices = self.pool(self.pad(f_maps))

            if self.return_pad:
                return output, indices, (coords[2], 0, coords[1], 0, coords[0], 0)
            else:
                return output, indices

        else:
            output = self.pool(self.pad(f_maps))

            if self.return_pad:
                return output, (coords[2], 0, coords[1], 0, coords[0], 0)
            else:
                return output


class CustomNetwork(nn.Module):
    def __init__(self):
        super(CustomNetwork, self).__init__()
        self.convolutions = nn.Sequential(
            nn.Conv3d(1, 8, 3, padding=1),
            # Size 8@30x40x30
            nn.BatchNorm3d(8),
            nn.LeakyReLU(),
            PadMaxPool3d(2, 2),
            # Size 8@15x20x15
            nn.Conv3d(8, 16, 3, padding=1),
            # Size 16@15x20x15
            nn.BatchNorm3d(16),
            nn.LeakyReLU(),
            PadMaxPool3d(2, 2),
            # Size 16@8x10x8)
            nn.Conv3d(16, 32, 3, padding=1),
            # Size 32@8x10x8
            nn.BatchNorm3d(32),
            nn.LeakyReLU(),
            PadMaxPool3d(2, 2),
            # Size 32@4x5x4
        )

        self.linear = nn.Sequential(nn.Dropout(p=0.5), nn.Linear(32 * 4 * 5 * 4, 2))

    def forward(self, x):
        x = self.convolutions(x)
        x = x.view(x.size(0), -1)
        x = self.linear(x)
        return x


def train(
    model: nn.Module,
    train_loader: DataLoader,
    criterion: nn.Module,
    optimizer,
    n_epochs: int,
) -> nn.Module:
    """Train a CNN.

    Parameters
    ----------
    model : nn.Module
        The neural network to train.

    train_loader : DataLoader
        A DataLoader wrapping a MRIDataset.

    criterion : nn.Module
        A method to compute the loss of a mini-batch of images.

    optimizer : torch.optim
        An optimization algorithm.

    n_epochs : int
        The number of epochs performed during training.

    Returns
    -------
    best_model : nn.Module
        The trained neural network.
    """
    from copy import deepcopy

    best_model = deepcopy(model)
    train_best_loss = np.inf

    for epoch in range(n_epochs):
        model.train()
        train_loader.dataset.train()
        for i, data in enumerate(train_loader, 0):
            # Retrieve mini-batch and put data on GPU with .cuda() is possible
            if torch.cuda.is_available():
                images, labels = data["image"].cuda(), data["label"].cuda()
            else:
                images, labels = data["image"], data["label"]
            # Forward pass
            outputs = model(images)
            # Loss computation
            loss = criterion(outputs, labels)
            # Back-propagation (gradients computation)
            loss.backward()
            # Parameters update
            optimizer.step()
            # Erase previous gradients
            optimizer.zero_grad()

        _, train_metrics = test(model, train_loader, criterion)

        print(
            f"Epoch {epoch}: loss = {train_metrics['mean_loss']:.4f}, "
            f"balanced accuracy = {train_metrics['balanced_accuracy']:.4f}"
        )

        if train_metrics["mean_loss"] < train_best_loss:
            best_model = deepcopy(model)
            train_best_loss = train_metrics["mean_loss"]

    return best_model


def test(model: nn.Module, data_loader: DataLoader, criterion: nn.Module) -> tuple:
    """Test a CNN.

    Parameters
    ----------
    model : nn.Module
        The neural network to test.

    data_loader : DataLoader
        A DataLoader wrapping a MRIDataset.

    criterion : nn.Module
        A method to compute the loss of a mini-batch of images.

    Returns
    -------
    results_df : pd.DataFrame
        The label predicted for every subject.

    results_metrics : dict
        A set of metrics.
    """
    model.eval()
    data_loader.dataset.eval()
    columns = ["participant_id", "proba0", "proba1", "true_label", "predicted_label"]
    results_df = pd.DataFrame(columns=columns)
    total_loss = 0

    with torch.no_grad():
        for i, data in enumerate(data_loader, 0):
            if torch.cuda.is_available():
                images, labels = data["image"].cuda(), data["label"].cuda()
            else:
                images, labels = data["image"], data["label"]
            outputs = model(images)
            loss = criterion(outputs, labels)
            total_loss += loss.item()
            probs = nn.Softmax(dim=1)(outputs)
            _, predicted = torch.max(outputs.data, 1)

            for idx, sub in enumerate(data["participant_id"]):
                row = [
                    sub,
                    probs[idx, 0].item(),
                    probs[idx, 1].item(),
                    labels[idx].item(),
                    predicted[idx].item(),
                ]
                row_df = pd.DataFrame([row], columns=columns)
                results_df = pd.concat([results_df, row_df])

    results_metrics = compute_metrics(
        results_df.true_label.values, results_df.predicted_label.values
    )
    results_df.reset_index(inplace=True, drop=True)
    results_metrics["mean_loss"] = total_loss / len(data_loader.dataset)

    return results_df, results_metrics


def compute_metrics(ground_truth, prediction) -> dict:
    """Computes the accuracy, sensitivity, specificity and balanced accuracy"""
    tp = np.sum((prediction == 1) & (ground_truth == 1))
    tn = np.sum((prediction == 0) & (ground_truth == 0))
    fp = np.sum((prediction == 1) & (ground_truth == 0))
    fn = np.sum((prediction == 0) & (ground_truth == 1))

    metrics_dict = dict()
    metrics_dict["accuracy"] = (tp + tn) / (tp + tn + fp + fn)
    metrics_dict["sensitivity"] = 0.0
    if tp + fn != 0:
        metrics_dict["sensitivity"] = tp / (tp + fn)
    metrics_dict["specificity"] = 0.0
    if fp + tn != 0:
        metrics_dict["specificity"] = tn / (fp + tn)

    metrics_dict["balanced_accuracy"] = (
        metrics_dict["sensitivity"] + metrics_dict["specificity"]
    ) / 2

    return metrics_dict



"""The data generation module."""

from ._toy_dataset import generate_toy_dataset
from ._cropped_hc_dataset import generate_cropped_hc_dataset

__all__ = ["generate_toy_dataset", "generate_cropped_hc_dataset"]



import torch
from pathlib import Path
from typing import Optional
from now_2023.utils import CropLeftHC, CropRightHC


def generate_cropped_hc_dataset(
    raw_dataset_folder: Path,
    hemi: str,
    output_folder: Optional[Path] = None,
    verbose: bool = True,
) -> None:
    """Generate a version of the cropped HC dataset.

    This function works with the raw dataset and will crop the available images.

    Parameters
    ----------
    raw_dataset_folder : Path
        The path to the raw dataset downloaded.

    hemi : str
        Either "left" or "right". Will crop the associated hipocampus.

    output_folder : Path, optional
        If specified, the output dataset will be written in this folder, under
        a "subjects" subfolder.
        Otherwise, the output dataset will be written in the same folder as the
        raw input dataset, under "cropped/subjects".

    verbose : bool, optional
        If True, the function will print information to stdout.
    """
    subjects_folder = raw_dataset_folder / "CAPS" / "subjects"
    if output_folder is None:
        output_folder = raw_dataset_folder / "cropped" / "subjects"
    else:
        output_folder = output_folder / "subjects"
    if verbose:
        print(f"Cropped images will be written in {output_folder}")
    if not output_folder.exists():
        output_folder.mkdir(parents=True)
    subjects = [_.name for _ in subjects_folder.iterdir()]

    for subject in subjects:
        if verbose:
            print(f"Generating HC ({hemi}) cropped images for subject {subject}...")
        input_image_folder = (
            subjects_folder
            / subject
            / "ses-M00"
            / "deeplearning_prepare_data"
            / "image_based"
            / "custom"
        )
        input_image_filename = f"{subject}_ses-M00_T1w_segm-graymatter_space-Ixi549Space_modulated-off_probability.pt"
        try:
            preprocessed_pt = torch.load(input_image_folder / input_image_filename)
        except:
            if verbose:
                print(
                    f"!!! Error reading input tensor for subject {subject}. Skipping..."
                )
            continue

        cropper = CropLeftHC(2) if hemi == "left" else CropRightHC(2)
        hc_cropped = cropper(preprocessed_pt)

        save_path = (
            output_folder
            / subject
            / "ses-M00"
            / "deeplearning_prepare_data"
            / "image_based"
            / "custom"
        )
        if not save_path.exists():
            save_path.mkdir(parents=True)
        if verbose:
            print(f"Saving {hemi} HC cropped image in {save_path}...")
        torch.save(
            hc_cropped,
            save_path
            / f"{subject}_ses-M00_T1w_segm-graymatter_space-Ixi549Space_modulated-off_probability_{hemi}.pt",
        )



"""The model module.

This module contains the machine learning and deep learning models used in the tutorial.
"""

from ._toy import ToyModel
from ._cnn import CNNModel

__all__ = ["CNNModel", "ToyModel"]



import argparse


if __name__ == "__main__":
    from ._toy_dataset import generate_toy_dataset

    parser = argparse.ArgumentParser()
    parser.add_argument("n_samples", help="The number of samples to generate.", type=int)
    parser.add_argument("output", help="Output file name.", type=str)
    args = parser.parse_args()
    generate_toy_dataset(n_samples=args.n_samples, output_filename=args.output)



def _dynamic_padding(number: int, width: int) -> str:
    return "sub-{number:0{width}d}".format(width=width, number=number)


def generate_toy_dataset(n_samples: int, output_filename: str) -> None:
    """Generate a toy dataset for classification.

    Parameters
    ----------
    n_samples : int
        The number of samples that the generated dataset should have.

    output_filename : str
        The path to the file in which the dataset should be written.
    """
    import pandas as pd
    from sklearn.datasets import make_classification

    X, y = make_classification(
        n_samples=n_samples,
        n_features=2,
        n_informative=2,
        n_redundant=0,
        n_repeated=0,
        n_classes=2,
        n_clusters_per_class=1,
        class_sep=2,
        flip_y=0.2,
        weights=[0.5, 0.5],
        random_state=17,
    )

    df = pd.DataFrame(X, columns=["HC_left_volume", "HC_right_volume"])
    df["group"] = y
    df["group"] = df["group"].apply(lambda x: "AD" if x == 1 else "CN")
    n_digits = len(str(n_samples))
    df["subject_id"] = [_dynamic_padding(x, n_digits) for x in range(1, len(df) + 1)]
    df.set_index("subject_id", inplace=True)
    df.to_csv(output_filename, sep="\t")



"""The utilities module."""

from ._statistics import characteristics_table
from ._images import CropLeftHC, CropRightHC


__all__ = [
    "CropLeftHC",
    "CropRightHC",
    "characteristics_table",
]



import pickle
import pandas as pd
import numpy as np
from sklearn import svm
import matplotlib.pyplot as plt
from typing import List


class ToyModel:
    """Toy model for classification."""

    def __init__(self, X: np.ndarray, y: np.ndarray):
        self.X = X
        self.y = y
        self.min_x = np.min(self.X[:, 0])
        self.max_x = np.max(self.X[:, 0])
        self.min_y = np.min(self.X[:, 1])
        self.max_y = np.max(self.X[:, 1])
        self.estimator = svm.SVC()

    @property
    def n_samples(self) -> int:
        """The number of samples."""
        return self.X.shape[0]

    @property
    def bbox(self) -> List[float]:
        return [self.min_x, self.max_x, self.min_y, self.max_y]

    @classmethod
    def from_dataframe(
        cls, df: pd.DataFrame, predictive_features: list[str], target_feature: str
    ):
        X = np.array(df[predictive_features])
        y = np.array([1 if x == "AD" else 0 for x in df[target_feature].values])
        return cls(X, y)

    def fit(self):
        self.estimator.fit(self.X, self.y)

    def save(self):
        with open("model.pkl", "wb") as fp:
            pickle.dump(self.estimator, fp)

    def plot(self):
        xx, yy = np.meshgrid(
            np.linspace(self.min_x, self.max_x, 500),
            np.linspace(self.min_y, self.max_y, 500),
        )
        Z = self.estimator.decision_function(np.c_[xx.ravel(), yy.ravel()])
        Z = Z.reshape(xx.shape)
        plt.imshow(
            Z,
            interpolation="nearest",
            extent=(xx.min(), xx.max(), yy.min(), yy.max()),
            aspect="auto",
            origin="lower",
            cmap=plt.cm.PuOr_r,
        )
        contours = plt.contour(xx, yy, Z, levels=[0], linewidths=2, linestyles="dashed")
        plt.scatter(
            self.X[:, 0],
            self.X[:, 1],
            s=30,
            c=self.y,
            cmap=plt.cm.Paired,
            edgecolors="k",
        )
        plt.xticks(())
        plt.yticks(())
        plt.title(f"Number of sample = {self.n_samples}")
        plt.axis(self.bbox)
        plt.show()



import random
from dataclasses import dataclass


@dataclass
class Point:
    x: int
    y: int
    z: int


class BaseCrop:
    def __init__(self, random_shift: int = 0):
        self.random_shift = random_shift
        self.train_mode = True
        self.bottom_left_point: Optional[Point] = None
        self.upper_right_point: Optional[Point] = None

    def _get_shift(self) -> int:
        """Return a random shift in train mode and no shift otherwise."""
        return (
            random.randint(-self.random_shift, self.random_shift)
            if self.train_mode
            else 0
        )

    def _get_slice(self, left_offset: int, right_offset: int):
        """Return a slice in a single dimension with possibly a random shift."""
        shift = self._get_shift()
        return slice(left_offset + shift, right_offset + shift)

    def _get_bbox(self) -> tuple:
        """Return the bounding box to use on the input image."""
        if self.bottom_left_point is None or self.upper_right_point is None:
            raise ValueError("No bounding box configured")
        return (
            Ellipsis,
            self._get_slice(self.bottom_left_point.x, self.upper_right_point.x),
            self._get_slice(self.bottom_left_point.y, self.upper_right_point.y),
            self._get_slice(self.bottom_left_point.z, self.upper_right_point.z),
        )

    def __call__(self, img):
        return img[self._get_bbox()].clone()

    def train(self):
        self.train_mode = True

    def eval(self):
        self.train_mode = False


class CropLeftHC(BaseCrop):
    """Crops the left hippocampus of a MRI non-linearly registered to MNI"""

    def __init__(self, random_shift: int = 0):
        super().__init__(random_shift=random_shift)
        self.bottom_left_point = Point(25, 50, 27)
        self.upper_right_point = Point(55, 90, 57)


class CropRightHC(BaseCrop):
    """Crops the right hippocampus of a MRI non-linearly registered to MNI"""

    def __init__(self, random_shift: int = 0):
        super().__init__(random_shift=random_shift)
        self.bottom_left_point = Point(65, 50, 27)
        self.upper_right_point = Point(95, 90, 57)



import torch
from pathlib import Path
from typing import Optional


def _show_slices(slices):
    """Function to display a row of image slices"""
    import matplotlib.pyplot as plt

    fig, axes = plt.subplots(1, len(slices))
    for i, slice in enumerate(slices):
        axes[i].imshow(slice.T, cmap="gray", origin="lower")


def plot_image(
    image: Path, cut_coords: tuple[int, int, int], title: Optional[str] = None
):
    """Plot a slice of a brain image.

    Paramaters
    ----------
    image : Path
        The path to the nifti image to be plotted.

    cut_coords : :obj:`tuple` of three :obj:`int`
        The coordinates of the slice to be displayed.

    title : str, optional
        The title of the plot.
    """
    import nibabel as nib
    import matplotlib.pyplot as plt
    from scipy.ndimage import rotate

    data = nib.load(image).get_fdata()
    slice_0 = data[:, :, cut_coords[0]]
    slice_1 = data[cut_coords[1], :, :]
    slice_2 = data[:, cut_coords[2], :]

    _show_slices([slice_0, rotate(slice_1, 90), rotate(slice_2, 90)])

    if title:
        plt.suptitle(title)

    plt.show()


def plot_tensor(
    tensor: Path | torch.Tensor,
    cut_coords: tuple[int, int, int],
    title: Optional[str] = None,
):
    """Plot a tensor image.

    Parameters
    ----------
    tensor : Path ot torch.Tensor
        The tensor image to be displayed. This can be a path to a ".pt" tensor file,
        or a torch.Tensor object.

    cut_coords : tuple of three ints
        The coordinates of the slice to be displayed.

    title : str, optional
        The title of the plot.
    """
    import torch
    import matplotlib.pyplot as plt

    if not isinstance(tensor, torch.Tensor):
        tensor = torch.load(tensor)
    slice_0 = tensor[0, cut_coords[0], :, :]
    slice_1 = tensor[0, :, cut_coords[1], :]
    slice_2 = tensor[0, :, :, cut_coords[2]]

    _show_slices([slice_0, slice_1, slice_2])

    if title:
        plt.suptitle(title)

    plt.show()


def plot_hc(dataset_folder: Path, subject: str, hemi: str, cut_coords: tuple = None):
    """Plot a brain image of the Hipocampus.

    This is a high-level function which assumes some structure in the input data.

    Parameters
    ----------
    dataset_folder : Path
        The path to the folder in which the raw input dataset of the tutorial was downloaded.

    subject : str
        The identifier of the subject for which a plot is desired.

    hemi : str
        Either "left" or "right". This indicates which hipocampus should be displayed.

    cut_coords : tuple of three ints
        The coordinates of the slice to be displayed.
    """
    image_folder = (
        dataset_folder
        / "subjects"
        / subject
        / "ses-M00"
        / "deeplearning_prepare_data"
        / "image_based"
        / "custom"
    )
    image_filename = f"{subject}_ses-M00_T1w_segm-graymatter_space-Ixi549Space_modulated-off_probability_{hemi}.pt"
    tensor = torch.load(image_folder / image_filename)
    plot_tensor(
        tensor,
        cut_coords=cut_coords or (15, 20, 15),
        title=f"Center slices of {hemi} HC of subject {subject}",
    )



import pandas as pd


def characteristics_table(df: pd.DataFrame, merged_df: pd.DataFrame) -> pd.DataFrame:
    """Creates a DataFrame that summarizes the characteristics of the DataFrame df.

    Parameters
    ----------
    df : pd.DataFrame
        The input dataframe on which to compute statistics.

    merged_df : pd.DataFrame

    Returns
    -------
    population_df : pd.DataFrame
        The dataframe representing the characteristic table.
    """
    import numpy as np

    diagnoses = np.unique(df.diagnosis.values)
    population_df = pd.DataFrame(
        index=diagnoses,
        columns=[
            "N",
            "age",
            "%sexF",
            "education",
            "MMS",
            "CDR=0",
            "CDR=0.5",
            "CDR=1",
            "CDR=2",
        ],
    )
    merged_df = merged_df.set_index(["participant_id", "session_id"], drop=True)
    df = df.set_index(["participant_id", "session_id"], drop=True)
    sub_merged_df = merged_df.loc[df.index]

    for diagnosis in population_df.index.values:
        diagnosis_df = sub_merged_df[df.diagnosis == diagnosis]
        population_df.loc[diagnosis, "N"] = len(diagnosis_df)
        # Age
        mean_age = np.mean(diagnosis_df.age_bl)
        std_age = np.std(diagnosis_df.age_bl)
        population_df.loc[diagnosis, "age"] = "%.1f ± %.1f" % (mean_age, std_age)
        # Sex
        population_df.loc[diagnosis, "%sexF"] = round(
            (len(diagnosis_df[diagnosis_df.sex == "F"]) / len(diagnosis_df)) * 100, 1
        )
        # Education level
        mean_education_level = np.nanmean(diagnosis_df.education_level)
        std_education_level = np.nanstd(diagnosis_df.education_level)
        population_df.loc[diagnosis, "education"] = "%.1f ± %.1f" % (
            mean_education_level,
            std_education_level,
        )
        # MMS
        mean_MMS = np.mean(diagnosis_df.MMS)
        std_MMS = np.std(diagnosis_df.MMS)
        population_df.loc[diagnosis, "MMS"] = "%.1f ± %.1f" % (mean_MMS, std_MMS)
        # CDR
        for value in ("0", "0.5", "1", "2"):
            population_df.loc[diagnosis, "CDR=%s" % value] = len(
                diagnosis_df[diagnosis_df.cdr_global == float(value)]
            )

    return population_df



import torch
import pandas as pd
from pathlib import Path

from torch.utils.data import Dataset, DataLoader, sampler
from typing import Optional, Callable
from dataclasses import dataclass


@dataclass
class Sample:
    """Class reprsenting a sample from an MRIDataset."""

    image: torch.Tensor
    label: str
    participant_id: str
    session_id: str
    hemi: str


class MRIDataset(Dataset):
    def __init__(
        self,
        img_dir: str,
        data_df: pd.DataFrame,
        transform: Optional[Callable] = None,
    ):
        """
        Parameters
        ----------
        img_dir : str
            Path to the CAPS directory containing preprocessed images.

        data_df : DataFrame
            Metadata of the population.
            Columns include `participant_id`, `session_id`, and `diagnosis`.

        transform : Callable, optional
            List of transforms applied on-the-fly, chained with torchvision.transforms.Compose.
        """
        self.img_dir: Path = Path(img_dir)
        self.transform = transform
        self.data_df = data_df
        self.label_code = {"AD": 1, "CN": 0}
        self.size = self[0]["image"].shape

    def __len__(self) -> int:
        return len(self.data_df)

    def __getitem__(self, idx: int) -> dict:
        return self._get_sample(idx).__dict__

    def _get_sample(self, idx: int) -> Sample:
        diagnosis = self.data_df.loc[idx, "diagnosis"]
        label = self.label_code[diagnosis]
        participant_id = self.data_df.loc[idx, "participant_id"]
        session_id = self.data_df.loc[idx, "session_id"]
        hemi = self.data_df.loc[idx, "hemi"]

        image_filename = f"{participant_id}_{session_id}_T1w_segm-graymatter_space-Ixi549Space_modulated-off_probability_{hemi}.pt"
        image_folder = (
            self.img_dir
            / "subjects"
            / participant_id
            / session_id
            / "deeplearning_prepare_data"
            / "image_based"
            / "custom"
        )
        image = torch.load(image_folder / image_filename)

        if self.transform:
            image = self.transform(image)

        return Sample(image, label, participant_id, session_id, hemi)

    def train(self):
        if self.transform:
            self.transform.train()

    def eval(self):
        if self.transform:
            self.transform.eval()
